#ifndef _KERNEL_INTERFACE_H_
#define _KERNEL_INTERFACE_H_

/***************************************************************************
 *
 * kernel_interface.h
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (C) 1999, 2000 NAI Labs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.  This program
 * is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.  You should have received a copy of the
 * GNU General Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 * This file is the public interface to kernel_interface.c
 *
 *
 ***************************************************************************/

#include <linux/sched.h>      /* for struct task_struct */
#include <linux/fs.h>         /* for struct inode and device/inode # macros */
#include <linux/dcache.h>     /* for struct dentry */
#include <linux/netdevice.h>  /* for struct device (the NIC structure) */

#include "lomac_level.h"

/* Conceptually, we map the LOMAC notion of subject onto the kernel's
 * notion of process group.  When intercepting system calls, the 
 * easiest method of determining which process group is responsible
 * for the call is to examine the `pgrp' field of the `current'
 * task_struct.  Since the `current' task_struct pointer is so convenient,
 * we define lomac_subject_t in terms of task_struct's rather than
 * than pgrp integers.
 */

typedef struct task_struct lomac_subject_t;

int  get_subject_id( const lomac_subject_t *p_subject, char *id_s,
		     int bufflen );
void set_subject_level( lomac_subject_t *p_subject, level_t level );
void get_subject_level( const lomac_subject_t *p_subject, level_t *p_level );


/* Early versions of LOMAC designed to work on 2.0 Linux kernels mapped
 * the "object" concept onto inodes (vnodes in BSD terms), because
 * the kernel used inodes to represent (just about?) everything LOMAC
 * considered an object.  Later versions of LOMAC designed to work on
 * 2.2 Linux kernels changed this arrangement by considering dentrys
 * to be objects, instead, because having dentry pointers instead of
 * inode pointers made it easy for LOMAC to use the kernel's d_path()
 * function to determine the pathnames of objects.  
 *
 * The object<->dentry arrangement isn't as neat as the original
 * object<->inode arrangement: although most dentrys contain a pointer
 * to an inode, the mapping between dentrys and inodes is not 1:1.
 * For example, although most dentrys point to exactly one inode (case
 * 1), some dentrys are placeholders for negative name lookups and
 * don't point to a inode (case 2), and where hard links are involved
 * some inodes are pointed to by multiple dentys (case 3).  LOMAC
 * concerns itself only with existing files (case 1).  It does not
 * make decisions about non-existent files, so dentrys without inodes
 * (case 2) aren't a problem.  However, LOMAC cannot overcome the
 * ambiguity of inodes with multiple dentries (case 3) - LOMAC's
 * problems with objects with multiple hard-linked names each
 * indicating a different level is a known and documented problem.
 *
 * In the Linux 2.4 kernels, the d_path() function requires a new
 * `struct vfsmount' * parameter.  The proper values for this parameter
 * are found in the `struct file' structure.  These structures point
 * to dentry structures, which point to inodes.  In order to get
 * easy access to the vfsmount pointers stored in the file structures,
 * the Linux 2.4 version of LOMAC makes "object" equivalent to `struct
 * file'.  This adds a little more complexity on top of the earlier
 * object<->dentry arrangement, since several file structures can
 * point to the same dentry.
 *
 */

typedef struct dentry lomac_object_t;

int  get_object_id( const lomac_object_t *p_object, char *id_s,
		    int bufflen );
int  get_object_canabspath( const lomac_object_t *p_object, char *path_s,
			    int bufflen );
void set_object_level( lomac_object_t *p_object, level_t level );
void get_object_level( const lomac_object_t *p_object, level_t *p_level );


/* These functions find objects based on certain criteria, such as  *
 * path.  They require the caller to invoke object_done() when they *
 * are finished with the found objects.                             */
lomac_object_t *get_subject_program_object( lomac_subject_t *p_subject );
lomac_object_t *path_to_object( const char *path_s );
void object_done( lomac_object_t *p_object );


/* Here are some types and routines to handle NICs.  LOMAC determines *
 * the level of network sockets according to the NIC they have read   *
 * packets from.                                                      */
typedef struct device lomac_nic_t;


/* This macro must take a string arg and print it to the system log. */
#define LOG( str )( printk( (str) ) )
/* hack - we use this prefix to set kernel log verbosity */
#define LOG_PREFIX KERN_INFO

/* This macro must print a message to the system console *
 * and panic the kernel.                                 */
#ifdef PARANOID
extern char *global_wrapper_name_s;
#define PANIC(s) printk( "LOMAC PANIC while handling %s\n", \
                         global_wrapper_name_s ); \
                 panic(s);
#else
#define PANIC panic
#endif /* if/else PARANOID */

/* These macro must present various libc str...(3) behaviors. */
#define STRCMP  strcmp
#define STRNCAT strncat
#define STRNLEN strnlen

/* This macro must place the decimal ASCII representation of *
 * `data' into `ascii'.                                      */
#define TOASCII( data, ascii )( sprintf( (ascii), "%u", (data) ) )

/* MAX_DIGITS is a sufficient number of decimal digits to describe a *
 * 64-bit uint with a null terminator, rounded to a multiple of 8.   */
#define MAX_DIGITS 24 

/* MAX_PATH_LEN is the length of the longest possible path string, *
 * plus 1 for a null terminator.                                   */
#define MAX_PATH_LEN PATH_MAX

/* LOG_BUFFER_LENGTH is a string length "large enough" to hold        *
 * interesting log messages.  You can set this to whatever you        *
 * want, the log routines will be careful not to exceed this limit.   *
 * Since some routines use log_s as scratch path space, this constant *
 * should always be at least MAX_PATH_LEN long.                       */
#define LOG_BUFFER_LENGTH MAX_PATH_LEN +1


#endif






