#ifndef _KERNEL_INTERFACE_H_
#define _KERNEL_INTERFACE_H_

/***************************************************************************
 *
 * kernel_interface.h
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (c) 1999-2001 NAI Labs, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * 
 *   Neither the name of NAI Labs, Inc. nor the names of its contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * This file is the public interface to kernel_interface.c
 *
 *
 ***************************************************************************/

#include <linux/sched.h>      /* for struct task_struct */
#include <linux/fs.h>         /* for struct inode and device/inode # macros */
#include <linux/dcache.h>     /* for struct dentry */
#include <linux/netdevice.h>  /* for struct device (the NIC structure) */

#include "lomac_level.h"

/* Conceptually, we map the LOMAC notion of subject onto the kernel's
 * notion of process group.  When intercepting system calls, the 
 * easiest method of determining which process group is responsible
 * for the call is to examine the `pgrp' field of the `current'
 * task_struct.  Since the `current' task_struct pointer is so convenient,
 * we define lomac_subject_t in terms of task_struct's rather than
 * than pgrp integers.
 */

typedef struct task_struct lomac_subject_t;

int  get_subject_id( const lomac_subject_t *p_subject, char *id_s,
		     int bufflen );
void set_subject_level( lomac_subject_t *p_subject, level_t level );
void get_subject_level( const lomac_subject_t *p_subject, level_t *p_level );


typedef struct dentry lomac_object_t;

int  get_object_id( const lomac_object_t *p_object, char *id_s,
		    int bufflen );
int  get_object_canabspath( const lomac_object_t *p_object, char *path_s,
			    int bufflen );
void set_object_level( lomac_object_t *p_object, level_t level );
void get_object_level( const lomac_object_t *p_object, level_t *p_level );


/* These functions find objects based on certain criteria, such as  *
 * path.  They require the caller to invoke object_done() when they *
 * are finished with the found objects.                             */
lomac_object_t *get_subject_program_object( lomac_subject_t *p_subject );
lomac_object_t *path_to_object( const char *path_s );
void object_done( lomac_object_t *p_object );


/* Here are some types and routines to handle NICs.  LOMAC determines *
 * the level of network sockets according to the NIC they have read   *
 * packets from.                                                      */
typedef struct device lomac_nic_t;


/* This macro must take a string arg and print it to the system log. */
#define LOG( str )( printk( (str) ) )
/* hack - we use this prefix to set kernel log verbosity */
#define LOG_PREFIX KERN_INFO

/* This macro must print a message to the system console *
 * and panic the kernel.                                 */
#ifdef PARANOID
extern char *global_wrapper_name_s;
#define PANIC(s) printk( "LOMAC PANIC while handling %s\n", \
                         global_wrapper_name_s ); \
                 panic(s);
#else
#define PANIC panic
#endif /* if/else PARANOID */

/* These macro must present various libc str...(3) behaviors. */
#define STRCMP  strcmp
#define STRNCAT strncat
#define STRNLEN strnlen

/* This macro must place the decimal ASCII representation of *
 * `data' into `ascii'.                                      */
#define TOASCII( data, ascii )( sprintf( (ascii), "%u", (data) ) )

/* MAX_DIGITS is a sufficient number of decimal digits to describe a *
 * 64-bit uint with a null terminator, rounded to a multiple of 8.   */
#define MAX_DIGITS 24 

/* MAX_PATH_LEN is the length of the longest possible path string, *
 * plus 1 for a null terminator.                                   */
#define MAX_PATH_LEN PATH_MAX

/* LOG_BUFFER_LENGTH is a string length "large enough" to hold        *
 * interesting log messages.  You can set this to whatever you        *
 * want, the log routines will be careful not to exceed this limit.   *
 * Since some routines use log_s as scratch path space, this constant *
 * should always be at least MAX_PATH_LEN long.                       */
#define LOG_BUFFER_LENGTH MAX_PATH_LEN +1


#endif






