/***************************************************************************
 *
 * test_lkm.c
 *
 * LOMAC - Low Water-Mark Mandatory Access Control for Linux 
 * Copyright (c) 1999-2001 NAI Labs, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * 
 *   Neither the name of NAI Labs, Inc. nor the names of its contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 *
 * Test lkm for kernel interface path-handling code.
 *
 ***************************************************************************/

#include <asm/uaccess.h>    /* for verify_area() */
#include <linux/module.h>   /* for module macros */
#include <linux/mm.h>       /* for __get_free_page() */

#include "kernel_util.h"

#ifdef SIMULATE_CHROOT
struct dentry *p_fake_root;
#endif
char *global_wrapper_name_s = "none";

/* Commands for ioctl() interface. */
#define GET_PID_LEVEL_IOCTL_CMD  0
#define GET_FILE_LEVEL_IOCTL_CMD 1


int do_ioctl( struct inode *, struct file *, unsigned int, unsigned long );
void init_driver( void );

/* This is the file operation jump table for the LOMAC *
 * character device.  We'll be using ioctl, only.      */
struct file_operations lomac_fops = { NULL, NULL, NULL, NULL, NULL,
				      do_ioctl, NULL, NULL, NULL, NULL,
				      NULL, NULL, NULL };

int lomac_major;           /* holds major number of lomac device */

/* do_ioctl()
 *
 * in:     p_inode_ignored - ignored
 *         p_file_ignored  - ignored
 *         command - GET_PID_LEVEL_IOCTL_CMD  - return level of pid in `arg'.
 *                   GET_FILE_LEVEL_IOCTL_CMD - return level of file named
 *                                              by the string `arg'.
 *         arg     - if `command' is GET_PID_LEVEL_IOCTL_CMD, then `arg' is
 *                   the pid of the process whose level we want to query.
 *                   if `command' is GET_FILE_LEVEL_IOCTL_CMD, then `cmd'
 *                   points to the name of the file whose level we want
 *                   to query, stored as a null-terminated string.
 * out:    nothing
 * return: the level of the process or file indicated by `command' and `arg',
 *         or -1 if no such process or file exists.  Also returns -1 on
 *         a bad `cmd' parameter.
 *
 *     User-space utility programs may query the state of the LOMAC LKM
 * through this ioctl() interface.
 *
 */

int do_ioctl( struct inode *p_inode_ignored, struct file *p_file_ignored,
	      unsigned int command, unsigned long arg ) {

  unsigned int filename_length;       /* length of file name from userspace */
  int ret_val;
  char *path_in_s;
  char *path_out_s;
  struct dentry *p_dentry;
  struct dentry *p_dir_dentry;
  char *chroot_s;

  printk( "in do_ioctl\n" );

  if( !( path_out_s = (char *)__get_free_page( GFP_KERNEL ) ) ) {
    return( -ENOMEM );
  }

  printk( "got memory\n" );
  
  switch( command ) {
  case GET_PID_LEVEL_IOCTL_CMD:          /* look up process level */
    return( -1 );
    break;

  case GET_FILE_LEVEL_IOCTL_CMD:         /* look up file level */

    printk( "handling file operation\n" );

#ifdef SIMULATE_CHROOT
    /* simulate chroot */
    current->fs->root = p_fake_root;
#endif

    /* When a file level is queried, we expect `arg' to point to a     *
     * variable-sized block of user-space memory organized as follows: *
     * First, an unsigned int containing the size of a following       *
     * buffer, and second, the buffer itself.  The buffer is used to   *
     * contain the absolute path to the file in question, as a         *
     * null-terminated string.                                         */

    /* First, verify that we can read the string-length uint, *
     * and copy it into kernel memory as `filename_length'.   */
    if( verify_area( VERIFY_READ, (void *)arg, sizeof( unsigned int ) ) ) {
      ret_val = -1;
      break;
    }
    copy_from_user( &filename_length, (void *)arg, sizeof( int ) );
    arg = (unsigned int)((unsigned int *)arg + 1);  /* skip over size uint */
    path_in_s = getname( (char *)arg );

    printk( "accomplished copyin\n" );

    p_dentry = p_dir_dentry = NULL;
    ret_val = make_canabspath( path_in_s, path_out_s, 
			       &chroot_s,
			       &p_dir_dentry, 
			       &p_dentry );

    printk( "cleaning up\n" );

    putname( path_in_s );
    if( p_dentry ) {
      dput( p_dentry );
    }
    if( !( ret_val ) ) {
      printk( "real: %s\n", path_out_s );
      printk( "chroot: %s\n", chroot_s );
    } else {
      printk( "no canabspath\n" );
    }

    if( p_dentry ) {
      printk( "file dentry set\n" );
    } else {
      printk( "file dentry clear\n" );
    }

    if( p_dir_dentry ) {
      printk( "dir dentry set\n" );
    } else {
      printk( "dir dentry clear\n" );
    }

    break;
    
  default:                               /* error - bad cmd value */

    ret_val = -1;

  } /* switch( command ) */

  free_page( (unsigned long)path_out_s );
  return( ret_val );

} /* do_ioctl() */



/* init_module()
 *
 * in:     nothing
 * out:    nothing
 * return: value      condition
 *         -----      ---------
 *         0          success
 *
 */

int init_module( void ) {

#ifdef SIMULATE_CHROOT
  p_fake_root = kernel_namei( "/home" );
#endif

  /* initialize driver (user-space info interface) */
  if( 0 > ( lomac_major = register_chrdev( 0, "lomac", &lomac_fops ) ) ) {
    panic( "LOMAC/K: no major device number available for lomac" );
  }

  /* Print a victory message to the console, indicating a successful load. */
  printk( "loaded test LKM\n" ); 
  return( 0 );

} /* init_module() */



/* cleanup_module()
 * 
 * in:     nothing
 * out:    nothing
 * return: nothing
 *
 * This function is called by the kernel in response to a request to unload
 * the lomac lkm.  Someday, this function should unpatch the system call
 * vector so the kernel can keep running once the LKM is gone, but this
 * requires some method to rescue processes that are currently blocked
 * in lomac-intercepted system calls.
 */

void cleanup_module( void ) {

#ifdef SIMULATE_CHROOT
  dput( p_fake_root );
#endif

  unregister_chrdev( lomac_major, "lomac" );
  printk( "unloaded test LKM\n" );

} /* cleanup_module() */





